@@ -9,6 +9,7 @@ import warnings
 from typing import Optional, Union, Callable
 
 import numpy as np
+from scipy.ndimage import gaussian_filter, gaussian_filter1d, zoom
 
 from astropy import units
 from astropy.io import fits
@@ -20,12 +21,19 @@ from astropy.nddata.utils import Cutout2D
 
 from .utils import get_pc_transform_params, rotate_data, get_pbar
 
+
 KNOWN_SPEC_EXT_NAMES = ['spec', 'spectrum', 'flux', 'data', 'sci', 'science']
 KNOWN_VARIANCE_EXT_NAMES = ['stat', 'stats', 'var', 'variance', 'noise', 'err']
 KNOWN_MASK_EXT_NAMES = ['mask', 'platemask', 'footprint', 'dq']
 KNOWN_RGB_EXT_NAMES = ['r', 'g', 'b', 'red', 'green', 'blue']
 
 
+def __simple_report_callback(k, total):
+    pbar = get_pbar(k / total)
+    sys.stderr.write(f"\r{pbar} {k} / {total} \r")
+    sys.stderr.flush()
+
+
 def __cutout_argshandler(options=None):
     """
     Parse the arguments given by the user.
@@ -126,6 +134,113 @@ def __cutout_argshandler(options=None):
     return args
 
 
+def __smoothing_argshandler(options=None):
+    """
+    Parse the arguments given by the user.
+
+    Inputs
+    ------
+    options: list or None
+        If none, args are parsed from the command line, otherwise the options
+        list is used as input for argument parser.
+
+    Returns
+    -------
+    args: Namespace
+        A Namespace containing the parsed arguments. For more information see
+        the python documentation of the argparse module.
+    """
+    parser = argparse.ArgumentParser(
+        description='Apply a gaussian smoothing kernel to a spectral cubes '
+                    'spatially and/or along the spectral axis.'
+    )
+
+    parser.add_argument(
+        'input_fits_files', metavar='INPUT_FIST', type=str, nargs='+',
+        help='The spectral cube (or image) from which to extract a cutout.'
+    )
+    parser.add_argument(
+        '--spatial-sigma', metavar='SIGMA', type=float, default=1.0,
+        help='Set the sigma for the spatial gaussian kernel. If %(metavar)s '
+        'is 0 then no spatial smoothing is applied. If not specified the '
+        'default value %(metavar)s=%(default)f is used.'
+    )
+    parser.add_argument(
+        '--spatial-supersample', metavar='ZOOM_FACTOR', type=int, default=0,
+        help='Set the spatial supersampling factor. If %(metavar)s <= 1 then '
+        'no supersampling is applied. %(metavar)s=2 means that the output cube'
+        ' will have a doubled width and height, and so on. The default value '
+        'is %(metavar)s=%(default)d.'
+    )
+    parser.add_argument(
+        '--wave-supersample', metavar='ZOOM_FACTOR', type=int, default=0,
+        help='Set the wavelength supersampling factor. If %(metavar)s <= 1 '
+        'then no supersampling is applied. %(metavar)s=2 means that the output'
+        'cube will have a doubled spectral resolution, and so on. '
+        'The default value is %(metavar)s=%(default)d.'
+    )
+    parser.add_argument(
+        '--wave-sigma', metavar='SIGMA', type=float, default=0,
+        help='Set the sigma for the spectral gaussian kernel. If %(metavar)s '
+        'is 0 then no spectral smoothing is applied. If not specified the '
+        'default value %(metavar)s=%(default)f is used.'
+    )
+    parser.add_argument(
+        '--info-hdu', metavar='INFO_HDU', type=int, default=0,
+        help='The HDU containing cube metadata. If this argument '
+        'Set this to -1 to automatically detect the HDU containing the info. '
+        'NOTE that this value is zero indexed (i.e. firts HDU has index 0).'
+    )
+
+    parser.add_argument(
+        '--spec-hdu', metavar='SPEC_HDU', type=int, default=-1,
+        help='The HDU containing the spectral data to use. If this argument '
+        'Set this to -1 to automatically detect the HDU containing spectra. '
+        'NOTE that this value is zero indexed (i.e. second HDU has index 1).'
+    )
+
+    parser.add_argument(
+        '--var-hdu', metavar='VAR_HDU', type=int, default=-1,
+        help='The HDU containing the variance of the spectral data. '
+        'Set this to -1 if no variance data is present in the cube. '
+        'The default value is %(metavar)s=%(default)s.'
+        'NOTE that this value is zero indexed (i.e. third HDU has index 2).'
+    )
+
+    parser.add_argument(
+        '--mask-hdu', metavar='MASK_HDU', type=int, default=-1,
+        help='The HDU containing the valid pixel mask of the spectral data. '
+        'Set this to -1 if no mask is present in the cube. '
+        'The default value is %(metavar)s=%(default)s.'
+        'NOTE that this value is zero indexed (i.e. fourth HDU has index 3).'
+    )
+
+    parser.add_argument(
+        '--verbose', action='store_true', default=False,
+        help="Print verbose outout."
+    )
+    args = None
+    if options is None:
+        args = parser.parse_args()
+    else:
+        args = parser.parse_args(options)
+
+    for inp_fits_file in args.input_fits_files:
+        if (inp_fits_file is not None) and (not os.path.isfile(inp_fits_file)):
+            print(f"The file {inp_fits_file} does not exist!")
+            sys.exit(1)
+
+    if (args.spatial_sigma == 0) and (args.wave_sigma == 0):
+        print(
+            "Spatial smoothing and spectral smooting cannot be both disabled, "
+            "please set at least one the options --wave-sigma or "
+            "--spatial-sigma."
+        )
+        sys.exit(1)
+
+    return args
+
+
 def parse_regionfile(regionfile: str, key_ra: str = 'RA', key_dec: str = 'DEC',
                      key_a: str = 'A_WORLD', key_b: str = 'B_WORLD',
                      key_theta: str = 'THETA_WORLD', key_wmin: str = 'WMIN',
@@ -340,8 +455,7 @@ def get_rgb_cutout(data: Union[tuple, list, np.ndarray],
     Parameters
     ----------
     data : np.ndarray or tuple or list
-        The actual image data. Should have only two dimensions (a grayscale
-        image has only X and Y corrdinates).
+        The actual image data.
     center : astropy.coordinate.SkyCoord or tuple.
         The center of the cutout. If a SkyCoord is provided then a WCS for the
         image data musto also be specified with the parameter data_wcs.
@@ -604,11 +718,11 @@ def _get_fits_data_structure(fits_file):
 
             # If cannot determine which extensions cointain data,
             # then just use the second extension
-            if data_structure['data-ext'] is None:
+            if not data_structure['data-ext']:
                 data_ext = f[1]
                 data_structure['data-ext'] = [1, ]
 
-        data_shape = data_ext.shape
+        data_shape = data_ext.data.shape
         if len(data_shape) == 2:
             # A 2D image, we should check other extensions to
             # determine if its an RGB multi-extension file
@@ -733,9 +847,198 @@ def get_hdu(hdl, valid_names, hdu_index=-1, msg_err_notfound=None,
     return valid_hdu
 
 
-def cutout_main(options=None):
+def cube_tiled_func(data, func, tile_size, *args, **kwargs):
+    data_shape = data.shape[-2:]
+    if isinstance(data, np.ma.MaskedArray):
+        result = np.ma.zeros(data_shape)
+    else:
+        result = np.zeros(data_shape)
+    for j in np.arange(data_shape[0], step=tile_size):
+        for k in np.arange(data_shape[1], step=tile_size):
+            tile = data[:, j:j+tile_size, k:k+tile_size]
+            # Skip empty tiles:
+            if not np.isfinite(tile).any():
+                result[j:j+tile_size, k:k+tile_size] = np.nan
+                try:
+                    result[j:j+tile_size, k:k+tile_size].mask = True
+                except AttributeError:
+                    pass
+                continue
+
+            processed_tile = func(tile, *args, **kwargs).copy()
+            result[j:j+tile_size, k:k+tile_size] = processed_tile
+            try:
+                result[j:j+tile_size, k:k+tile_size].mask = processed_tile.mask
+            except AttributeError:
+                pass
+
+    return result
+
+
+def correlate_spaxel(cube_data: np.ndarray,
+                     spaxel_data: np.ndarray,
+                     similarity_function: Optional[str] = 'rms'):
+
+    if spaxel_data.shape[0] != cube_data.shape[0]:
+        raise ValueError(
+            "spaxel_data and cube_data must have the same first dimension."
+        )
+
+    if len(spaxel_data.shape) == 1:
+        spaxel_data = spaxel_data[:, None, None]
+
+    x = cube_data - np.nanmedian(cube_data, axis=0)
+    x = x / np.nanmax(x, axis=0)
+    y = spaxel_data - np.nanmedian(spaxel_data)
+    y = y / np.nanmax(y)
+
+    if similarity_function == 'rms':
+        res = np.sqrt(np.nanmean((x - y)**2, axis=0))
+        return 1/(1 + res)
+    elif similarity_function == 'correlation':
+        res = np.nansum(x * y, axis=0)
+        return res / (np.nansum(x**2, axis=0) * np.nansum(y**2))
+
+
+def self_correlate(data: np.ndarray,
+                   data_mask: Optional[np.ndarray] = None,
+                   similarity_sigma_threshold: Optional[float] = 5,
+                   tile_size: Optional[int] = 32,
+                   block_size: Optional[int] = 2,
+                   similarity_function: Optional[str] = 'rms',
+                   report_callback: Optional[Callable] = None) -> np.ndarray:
+    if data_mask is not None and data.shape != data_mask.shape:
+            raise ValueError("data and data_mask must have the same shape!")
+
+    hei, wid = data.shape[1:]
+
+    sim_table = np.zeros((hei, wid))
+    # Dor each spaxel in the cube
+    block_id = 0
+    for h in np.arange(hei, step=block_size):
+        for k in np.arange(wid, step=block_size):
+            block_id += 1
+            if (
+                (sim_table[h:h+block_size, k:k+block_size] != 0).any() or
+                (
+                    data_mask is not None and
+                    (data_mask[:, h:h+block_size, k:k+block_size] != 0).any()
+                )
+            ):
+                continue
+
+            if report_callback is not None:
+                report_callback(block_id + 1, int(wid*hei / (block_size**2)))
+
+            spaxel_data = np.nansum(
+                data[:, h:h+block_size, k:k+block_size],
+                axis=(1, 2)
+            )
+
+            if not np.isfinite(spaxel_data).any():
+                continue
+
+            similarity_map = cube_tiled_func(
+                data,
+                correlate_spaxel,
+                tile_size=tile_size,
+                spaxel_data=spaxel_data,
+                similarity_function=similarity_function
+            )
+
+            thresh = np.nanmean(similarity_map)
+            thresh += similarity_sigma_threshold * np.nanstd(similarity_map)
+            similarity_mask = similarity_map >= thresh
+            sim_table[similarity_mask] = block_id
+    return sim_table
+
+
+def smooth_cube(data: np.ndarray, data_mask: Optional[np.ndarray] = None,
+                spatial_sigma: Optional[float] = 1.0,
+                wave_sigma: Optional[float] = 0.0,
+                report_callback: Optional[Callable] = None) -> np.ndarray:
     """
-    Run the main program.
+    Smooth a datacube spatially and/or along the spectral axis.
+
+    Parameters
+    ----------
+    data : numpy.ndarray
+        The spectral datacube.
+    data_mask : numpy.ndarray, optional
+        The mask for the spectral datacube. The default is None.
+    spatial_sigma : float, optional
+        The sigma for the spatial smoothing gaussian kernel.
+        The default is 1.0.
+    wave_sigma : float, optional
+        The sigma fot the spectral smoothing gaussian kernel.
+        The default is 0.0.
+    report_callback : Callable or None, optional
+        A callable that will be execute every time the cutout of a single
+        slice of the cube is computed. Must accept in input two arguments:
+            - the number of slice processed so far
+            - the total number of slices.
+
+    Raises
+    ------
+    ValueError
+        If the shape of data does not match the shape of data_mask.
+
+    Returns
+    -------
+    smoothed_arr : numpy.ndarray
+        The smoothed version of the input data.
+
+    """
+    if data_mask is not None and data.shape != data_mask.shape:
+            raise ValueError("data and data_mask must have the same shape!")
+
+    smoothed_arr = data.copy()
+
+    if wave_sigma > 0:
+        for h in range(smoothed_arr.shape[1]):
+            if report_callback is not None:
+                report_callback(h, smoothed_arr.shape[1])
+            for k in range(smoothed_arr.shape[2]):
+                smoothed_spaxel = gaussian_filter1d(
+                    smoothed_arr[:, h, k],
+                    sigma=wave_sigma,
+                    mode='constant'
+                )
+                smoothed_arr[:, h, k] = smoothed_spaxel
+
+    if report_callback is not None:
+        print("")
+
+    if spatial_sigma > 0:
+        for k, data_slice in enumerate(smoothed_arr):
+            if report_callback is not None:
+                report_callback(k + 1, smoothed_arr.shape[0])
+            smoothed_slice = gaussian_filter(
+                data_slice,
+                sigma=spatial_sigma,
+                mode='constant'
+            )
+            smoothed_arr[k] = smoothed_slice
+
+    if report_callback is not None:
+        print("")
+
+    if data_mask is not None:
+        smoothed_mask = data_mask.copy().astype(bool)
+        for k, mask_slice in enumerate(smoothed_mask):
+            if report_callback is not None:
+                report_callback(k + 1, smoothed_mask.shape[0])
+                smoothed_mask[k] &= ~np.isfinite(smoothed_arr[k])
+        if report_callback is not None:
+            print("")
+        smoothed_mask = smoothed_mask.astype('int8')
+
+    return smoothed_arr, smoothed_mask
+
+
+def smoothing_main(options=None):
+    """
+    Run the main cutout program.
 
     Parameters
     ----------
@@ -747,11 +1050,103 @@ def cutout_main(options=None):
     None.
 
     """
-    def simple_report_callback(k, total):
-        pbar = get_pbar(k / total)
-        sys.stderr.write(f"\r{pbar} {k} / {total} \r")
-        sys.stderr.flush()
+    args = __smoothing_argshandler(options)
 
+    if args.verbose:
+        report_callback = __simple_report_callback
+    else:
+        report_callback = None
+
+    for target_data_file in args.input_fits_files:
+        fits_base_name = os.path.basename(target_data_file)
+        fits_base_name = os.path.splitext(fits_base_name)[0]
+        if args.verbose:
+            print(f"\n[{fits_base_name}]")
+
+        with fits.open(target_data_file, mode='readonly') as hdul:
+                spec_hdu = get_hdu(
+                    hdul,
+                    hdu_index=args.spec_hdu,
+                    valid_names=KNOWN_SPEC_EXT_NAMES,
+                    msg_err_notfound=(
+                        "ERROR: Cannot determine which HDU contains spectral "
+                        "data, try to specify it manually!"
+                    ),
+                    msg_index_error="ERROR: Cannot open HDU {} to read specra!"
+                )
+
+                spec_wcs = WCS(spec_hdu.header)
+
+                var_hdu = get_hdu(
+                    hdul,
+                    hdu_index=args.var_hdu,
+                    valid_names=KNOWN_VARIANCE_EXT_NAMES,
+                    msg_err_notfound=(
+                        "WARNING: Cannot determine which HDU contains the "
+                        "variance data, try to specify it manually!"
+                    ),
+                    msg_index_error="WARNING: Cannot open HDU {} to read the "
+                                    "variance!",
+                    exit_on_errors=False
+                )
+
+                mask_hdu = get_hdu(
+                    hdul,
+                    hdu_index=args.mask_hdu,
+                    valid_names=KNOWN_MASK_EXT_NAMES,
+                    msg_err_notfound=(
+                        "WARNING: Cannot determine which HDU contains the "
+                        "mask data, try to specify it manually!"
+                    ),
+                    msg_index_error="WARNING: Cannot open HDU {} to read the "
+                                    "mask!",
+                    exit_on_errors=False
+                )
+
+                if mask_hdu is not None:
+                    data_mask = mask_hdu.data
+
+                if args.verbose:
+                    print(">>> applying smoothing...")
+                    print(f"  - spatial_sigma: {args.spatial_sigma}")
+                    print(f"  - wave_sigma: {args.wave_sigma}")
+
+                smoothed_spec, smoothed_mask = smooth_cube(
+                    data=spec_hdu.data,
+                    data_mask=data_mask,
+                    spatial_sigma=args.spatial_sigma,
+                    wave_sigma=args.wave_sigma,
+                    report_callback=report_callback
+                )
+
+                spec_hdu.data = smoothed_spec
+                if mask_hdu is not None:
+                    mask_hdu.data = smoothed_mask
+
+                out_fname = f"{fits_base_name}_smoothed.fits"
+                if args.verbose:
+                    print("  - saving to {out_fname}...")
+
+                hdul.writeto(
+                    out_fname,
+                    overwrite=True
+                )
+
+
+def cutout_main(options=None):
+    """
+    Run the main cutout program.
+
+    Parameters
+    ----------
+    options : list or None, optional
+        A list of cli input prameters. The default is None.
+
+    Returns
+    -------
+    None.
+
+    """
     def updated_wcs_cutout_header(orig_header, cutout_header):
         new_header = orig_header.copy()
 
@@ -832,7 +1227,7 @@ def cutout_main(options=None):
             f" DQ EXT: {data_structure['mask-ext']}\n",
             file=sys.stderr
         )
-        report_callback = simple_report_callback
+        report_callback = __simple_report_callback
     else:
         report_callback = None
 
@@ -1042,7 +1437,3 @@ def cutout_main(options=None):
                     f"WARNING: not implemente yet [{data_structure['type']}]!",
                     file=sys.stderr
                 )
-
-
-if __name__ == '__main__':
-    cutout_main()
